using System;
using System.Resources;
using System.Reflection;
using System.Xml.XPath;
using System.Xml.Xsl;
using System.Xml.Schema;
using System.Xml;
using System.IO;
using System.Text.RegularExpressions;
using System.Net;
using System.Text;
using GotDotNet.Exslt;
using GotDotNet.XInclude;

namespace NXslt {
    /// <summary>
    /// .NET XSLT Command line utility main class.
    /// <p>Author: Oleg Tkachenko, <address>olegt@multiconn.com</address></p>
    /// </summary>
    class NXslt : PerfTest {
        //Resource manager for l10n.        
        ResourceManager rm;
        //Parsed command line options
        NXsltOptions options;
		//Console.Error alias
		TextWriter stderr = Console.Error;

        /// <summary>
        /// Return values.
        /// </summary>
        internal enum ReturnValue {Error=-1, Ok=0};
                        
        public NXslt() {
            rm = new ResourceManager("NXslt.NXslt", 
                Assembly.GetExecutingAssembly());
        }               
        
        /// <summary>
        /// Prints the utility usage.
        /// </summary>
        /// <param name="rm">Resource manager for retrieving localized messages.</param>
        public static void PrintUsage(ResourceManager rm) {            
            Console.Error.WriteLine(rm.GetString("Usage.Header"), System.Environment.Version.Major, System.Environment.Version.Minor);
            Console.Error.WriteLine();
            Console.Error.WriteLine(rm.GetString("Usage.Body"));
        }
        
        /// <summary>
        /// Prints command line parsing error.
        /// </summary>
        /// <param name="rm">Resource manager for retrieving localized messages.</param>
        public static void PrintCommandLineParsingError(ResourceManager rm) {
            PrintUsage(rm);
            Console.Error.WriteLine(rm.GetString("Error.CommandLine.Parsing"));
            Console.Error.WriteLine();            
        }
        
        /// <summary>
        /// Prints command line parsing error.
        /// </summary>
        /// <param name="rm">Resource manager for retrieving localized messages.</param>        
        /// <param name="msgKey">Error message key.</param>
        public static void PrintCommandLineParsingError(ResourceManager rm, string msgKey) {
            PrintUsage(rm);
            Console.Error.WriteLine(rm.GetString("Error.CommandLine.Parsing"));
            Console.Error.WriteLine();            
            Console.Error.WriteLine(rm.GetString(msgKey));
        }
        
        /// <summary>
        /// The main entry point of the utility.
        /// </summary>
        [STAThread]
        static int Main(string[] args) {
			try {				
				NXslt nxslt = new NXslt();
				try {
					nxslt.options = NXsltArgumentsParser.ParseArguments(args, nxslt.rm);
				} catch (Exception e) {
					PrintCommandLineParsingError(nxslt.rm);				
					Console.Error.WriteLine(e);
					return (int)ReturnValue.Error;
				}
				if (nxslt.options == null)
					return (int)ReturnValue.Error;
				else
					return (int)nxslt.Process();			
			} catch (Exception e) {
				//This should catch everything
				Console.Error.WriteLine();				
				Console.Error.WriteLine("Some error has occured:");
				Console.Error.WriteLine(e);
				return (int)ReturnValue.Error;
			}
        }
               
        /// <summary>
        /// Process command line arguments and applies the specified stylesheet
        /// to the specified source document.
        /// </summary>        
        /// <returns>Return value code, 0 - Ok, others - error.</returns>       
        internal ReturnValue Process() {
            //Just show help
            if (options.ShowHelp) {
                PrintUsage(rm);
                return ReturnValue.Ok;
            }
            //Check that everything is in place
            if (options.Source == null && !options.LoadSourceFromStdin) {
                PrintCommandLineParsingError(rm, "Error.Missing.Source");
                return ReturnValue.Error;
            }
            if (options.Stylesheet == null && !options.LoadStylesheetFromStdin
                    && !options.GetStylesheetFromPI) {
                PrintCommandLineParsingError(rm, "Error.Missing.Stylesheet");                
                return ReturnValue.Error;
            }
            XPathDocument doc;                        
            float[] timings = new float[4];            
            XmlValidatingReader sourceReader, styleReader = null;
            XPathNavigator styleNavigator = null;
            //Load and parse source
            try {                                
                if (options.ShowTiming)                    
                    this.Start();
                if (options.LoadSourceFromStdin)
                    //Get source from stdin {
                    sourceReader = new XmlValidatingReader(new XmlTextReader(System.Console.OpenStandardInput()));                     
                else
                    //Get source from URI
                    sourceReader = new XmlValidatingReader(new XmlTextReader(options.Source));
                sourceReader.ValidationType = 
                    options.ValidateDocs? ValidationType.Auto : ValidationType.None;
                if (!options.ResolveExternals)
                    sourceReader.XmlResolver = null;				
                doc = new XPathDocument(
					options.ProcessXInclude? (XmlReader)new XIncludingReader(sourceReader) : sourceReader, 
                    options.StripWhiteSpace? XmlSpace.None: XmlSpace.Preserve);
                //Save source loading time
                if (options.ShowTiming) 
                    timings[0] = this.Stop();                                                
            } catch (XmlSchemaException xe) {
                ReportValidateError(xe);
                return ReturnValue.Error;            
            } catch (XmlException xe) {
                ReportParsingError(xe, options.LoadSourceFromStdin? 
                    rm.GetString("From.Stdin"):options.Source);
                return ReturnValue.Error;                                    
            } catch (Exception e) {
                //Another exception
                stderr.WriteLine();
                stderr.WriteLine(rm.GetString("Error.Loading.Doc"), options.Source);
                stderr.WriteLine();
                stderr.WriteLine(e);                
                return ReturnValue.Error;
            }        
            //Load, parse and compile stylesheet    
			ExsltTransform xslt = null;
			object extXslt = null;
            try {                
                if (options.ShowTiming)                    
                    this.Start();                
                if (options.GetStylesheetFromPI) {
                    //Get stylesheet URI from PI
                    XPathNavigator nav = doc.CreateNavigator();
                    options.Stylesheet = GetStylesheetURIFromPI(nav);
                    if (options.Stylesheet == null)
                        return ReturnValue.Error;
                    if (options.Stylesheet.StartsWith("#")) {
                        //Fragment identifier - embedded stylesheet
                        string id = options.Stylesheet.Remove(0, 1);
                        XPathNodeIterator ni = nav.Select("id('"+id+"')|/descendant::*[@id='"+id+"']");
                        if (ni.MoveNext()) {
                            styleNavigator = ni.Current;                            
                        } else {
                            //Unresolvable stylesheet URI
                            stderr.WriteLine();
							stderr.WriteLine(rm.GetString("Error.PI.Stylesheet.Not.Found"), options.Stylesheet);
							stderr.WriteLine();
                            return ReturnValue.Error;
                        }
                    } else {
                        //Lets resolve this value
                        XmlUrlResolver resolver = new XmlUrlResolver();
                        Uri fullUri = resolver.ResolveUri(new Uri(sourceReader.BaseURI), 
                            options.Stylesheet);
                        if (fullUri == null) {
                            //Unresolvable stylesheet URI
                            stderr.WriteLine();
							stderr.WriteLine(rm.GetString("Error.PI.Stylesheet.Not.Found"), fullUri);
							stderr.WriteLine();
                            return ReturnValue.Error;
                        }
                        options.Stylesheet = fullUri.AbsoluteUri;
                        styleReader = new XmlValidatingReader(new XmlTextReader(options.Stylesheet));
                    }                                        
                } else if (options.LoadStylesheetFromStdin)
                    //Read stylesheet from stdin
                    styleReader = new XmlValidatingReader(new XmlTextReader(System.Console.OpenStandardInput()));
                else
                    //Read stylesheet from URI
                    styleReader = new XmlValidatingReader(new XmlTextReader(options.Stylesheet));
                if (styleReader != null) {                    
                    styleReader.ValidationType = options.ValidateDocs? ValidationType.Auto : ValidationType.None;
                    if (!options.ResolveExternals)
                        styleReader.XmlResolver = null;
                }        
                //Save stylesheet loading time
                if (options.ShowTiming)
                    timings[1] = this.Stop();
				if (options.ExsltLibPath == null)
					xslt = new ExsltTransform();
				else {
					//Load EXSLT.NET assembly
					Assembly exsltAsmb = null;
					try {
						if (options.ExsltLibPath == "#GAC") {
							exsltAsmb = Assembly.LoadWithPartialName("GotDotNet.Exslt");
							if (exsltAsmb == null)
								throw new ApplicationException("GotDotNet.Exslt assembly not found in the GAC.");								
						} else
							exsltAsmb = Assembly.LoadFrom(options.ExsltLibPath);
					} catch (Exception e) {
						stderr.WriteLine();
						stderr.WriteLine(rm.GetString("Error.Load.ExsltLibNotFound"), 
							options.ExsltLibPath, e.Message);
						stderr.WriteLine();
						return ReturnValue.Error;
					}
					Type exsltTransformType = null;
					try {
						exsltTransformType = exsltAsmb.GetType("GotDotNet.Exslt.ExsltTransform", true);
					} catch (Exception e) {
						stderr.WriteLine();
						stderr.WriteLine(rm.GetString("Error.Load.ExsltTransformTypeNotFound"), 
							options.ExsltLibPath, e.Message);
						stderr.WriteLine();
						return ReturnValue.Error;
					}
					try {
						extXslt = exsltTransformType.GetConstructor(Type.EmptyTypes).Invoke(null);
					} catch (Exception e) {
						stderr.WriteLine();
						stderr.WriteLine(rm.GetString("Error.Load.ExsltTransformTypeCannotBeInstantiated"), 
							options.ExsltLibPath, e.Message);
						stderr.WriteLine();
						return ReturnValue.Error;
					}														
				}
                XmlResolver customResolver = null;                
                //Instantiate specified URI resolver
                if (options.ResolverTypeName != null) {                
                    Type resolverType = FindType(options, rm, options.ResolverTypeName);
                    if (resolverType == null)
                        return ReturnValue.Error;                        
                    try {
                        object o = resolverType.GetConstructor(Type.EmptyTypes).Invoke(null);
                        if (o is XmlResolver)
                            customResolver = o as XmlResolver;
                        else {
                            //Type is not XmlResolver
                            stderr.WriteLine();
                            stderr.WriteLine(rm.GetString("Error.Type.Not.XmlResolver"), options.ResolverTypeName);
                            stderr.WriteLine();
                            return ReturnValue.Error;
                        }
                    } catch (Exception e) {
                        //Type cannot be instantiated
                        stderr.WriteLine();
                        stderr.WriteLine(rm.GetString("Error.Create.Resolver"), options.ResolverTypeName, e);
                        stderr.WriteLine();
                        return ReturnValue.Error;                                                    
                    }
                }                                
                if (options.ShowTiming)                    
                    this.Start();            
                if (options.ExsltLibPath == null) {                        
					if (customResolver != null) {
						xslt.XmlResolver = customResolver;
						if (styleReader != null)											
							xslt.Load(
								options.ProcessXInclude? (XmlReader)new XIncludingReader(styleReader) : styleReader, 
								customResolver);
						else
							xslt.Load(styleNavigator, customResolver);
					} else {
						if (styleReader != null)					
							xslt.Load(options.ProcessXInclude? (XmlReader)new XIncludingReader(styleReader) : styleReader);
						else
							xslt.Load(styleNavigator);
					}
				} else {
					if (customResolver != null) {
						extXslt.GetType().GetProperty("XmlResolver").SetValue(extXslt, customResolver, null);						
						if (styleReader != null)											
							extXslt.GetType().GetMethod("Load", new Type[] {typeof(XmlReader), typeof(XmlResolver)}).Invoke(extXslt, 
								new Object[] {
									options.ProcessXInclude? (XmlReader)new XIncludingReader(styleReader) : styleReader, 
									customResolver
								});							
						else
							extXslt.GetType().GetMethod("Load", new Type[] {typeof(XPathNavigator), typeof(XmlResolver)}).Invoke(extXslt, 
								new Object[] {styleNavigator, customResolver});
					} else {
						if (styleReader != null) {					
							XmlReader r = options.ProcessXInclude? (XmlReader)new XIncludingReader(styleReader) : styleReader;
							MethodInfo mi = extXslt.GetType().GetMethod("Load", new Type[] {typeof(XmlReader)});
							mi.Invoke(extXslt, new Object[] {r});
						} else
							extXslt.GetType().GetMethod("Load", new Type[] {typeof(XPathNavigator)}).Invoke(extXslt, 
								new Object[] {styleNavigator});
					}
				}
                //Save stylesheet compilation time
                if (options.ShowTiming)
                    timings[2] = this.Stop();
            } catch (XmlSchemaException xe) {
                ReportValidateError(xe);                
                return ReturnValue.Error;            
            } 
            catch (XmlException xe) {
                ReportParsingError(xe, options.LoadStylesheetFromStdin? 
                    rm.GetString("From.Stdin"):options.Stylesheet);
                return ReturnValue.Error;
            } catch (Exception e) {
                stderr.WriteLine();
                stderr.WriteLine(rm.GetString("Error.Compile.Style"), options.Stylesheet);
                stderr.WriteLine();
                stderr.WriteLine(e);    
                return ReturnValue.Error;
            }            
            //Transform            
            try {    
                if (options.ShowTiming)                    
                    this.Start();
                if (options.OutFile != null) {
                    //Transform to a file
                    FileStream fs;
                    try {
                        fs = File.Create(options.OutFile);                          
                    } catch (Exception e) {
                        stderr.WriteLine();
                        stderr.WriteLine(rm.GetString("Error.Creating.File"), options.OutFile);
                        stderr.WriteLine();
                        stderr.WriteLine(e);
                        return ReturnValue.Error;        
                    }                                    
                    if (options.MultiOutput == true) {                        
                        MultiXmlTextWriter multiWriter = new MultiXmlTextWriter(fs, Encoding.UTF8);                        
                        DirectoryInfo dir = Directory.GetParent(options.OutFile);                       
                        Directory.SetCurrentDirectory(dir.ToString());
                        if (options.ExsltLibPath == null)
							xslt.Transform(doc, options.XslArgList, multiWriter);                        
						else
							extXslt.GetType().GetMethod("Transform", 
								new Type[] {typeof(XPathDocument), typeof(XsltArgumentList), typeof(XmlWriter)}).Invoke(extXslt, 
								new object[] {doc, options.XslArgList, multiWriter});
                    } else
						if (options.ExsltLibPath == null)
							xslt.Transform(doc, options.XslArgList, fs);
						else 
							extXslt.GetType().GetMethod("Transform",
							new Type[] {typeof(XPathDocument), typeof(XsltArgumentList), typeof(Stream)}).Invoke(extXslt, 
								new object[] {doc, options.XslArgList, fs});
                    fs.Close();
                } else {
                    //Transform to console                    
                    if (options.MultiOutput == true) {
                        MultiXmlTextWriter multiWriter = new MultiXmlTextWriter(Console.Out);
                        if (options.ExsltLibPath == null)
							xslt.Transform(doc, options.XslArgList, multiWriter);                        
						else
							extXslt.GetType().GetMethod("Transform",
							new Type[] {typeof(XPathDocument), typeof(XsltArgumentList), typeof(XmlWriter)}).Invoke(extXslt, 
								new object[] {doc, options.XslArgList, multiWriter});
                    } else
						if (options.ExsltLibPath == null)
							xslt.Transform(doc, options.XslArgList, Console.OpenStandardOutput());
						else
							extXslt.GetType().GetMethod("Transform", 
							new Type[] {typeof(XPathDocument), typeof(XsltArgumentList), typeof(Stream)}).Invoke(extXslt, 
								new object[] {doc, options.XslArgList, Console.OpenStandardOutput()});
                }
            } catch(Exception xe) {
                //Exception during XSL Transformation
                stderr.WriteLine();
                stderr.WriteLine(rm.GetString("Error.Transform"), options.Stylesheet);
                stderr.WriteLine();
                stderr.WriteLine(xe);
				return ReturnValue.Error;
            }            
            //Save transformation time and show timings report
            if (options.ShowTiming) {
                timings[3] = this.Stop();
                stderr.WriteLine(rm.GetString("Usage.Header"));
                stderr.WriteLine();
                stderr.WriteLine(rm.GetString("Timings"), timings[0], 
                    timings[1], timings[2], timings[3]);                
            }
            return ReturnValue.Ok;
        }
        
        /// <summary>
        /// Search stylesheet URI in xml-stylesheet PI.
        /// </summary>
        /// <param name="nav">XPath navigator over source document.</param>
        /// <returns>Stylesheet URI or null.</returns>
        internal string GetStylesheetURIFromPI(XPathNavigator nav) {             
            //Select the PI
            XPathNodeIterator ni = nav.Select("processing-instruction('xml-stylesheet')");
            if (ni.MoveNext()) {                        
                //Found PI node, look for href pseudo attribute
                Regex r = new Regex(@"href[ \n\t\r]*=[ \n\t\r]*""(.*)""|href[ \n\t\r]*=[ \n\t\r]*'(.*)'");
                Match m = r.Match(ni.Current.Value);
                if (m.Success)
                    //Found href pseudo attribute value
                    return m.Groups[1].Success? m.Groups[1].Value : 
                        m.Groups[2].Value;                                    
                else {
                    //Absent href preudo attribute
                    stderr.WriteLine();
                    stderr.WriteLine(rm.GetString("Error.Invalid.PI"));
                    stderr.WriteLine();
                    return null;
                }
            } 
            else {
                //Absent xml-stylesheet PI
                stderr.WriteLine();
				stderr.WriteLine(rm.GetString("Error.Invalid.PI"));
				stderr.WriteLine();
                return null;
            }                                
        } // GetStylesheetURIFromPI method   
        
        /// <summary>
        /// Reports validation error.
        /// </summary>
        /// <param name="xe">XmlSchemaException object.</param>
        internal void ReportValidateError(XmlSchemaException xe) {            
            stderr.WriteLine();
            stderr.WriteLine(rm.GetString("Error.Parsing.Doc"));
            stderr.WriteLine();
            stderr.WriteLine(rm.GetString("URL") + xe.SourceUri);                
            stderr.WriteLine(rm.GetString("Line") + xe.LineNumber);
            stderr.WriteLine(rm.GetString("Column") + xe.LinePosition);
            stderr.WriteLine(xe);    
        }
        
        /// <summary>
        /// Reports parsing error.
        /// </summary>
        /// <param name="xe">XmlException object.</param>
        /// <param name="uri">Source URI.</param>
        internal void ReportParsingError(XmlException xe, string uri) {                        
            stderr.WriteLine();
            stderr.WriteLine(rm.GetString("Error.Parsing.Doc"));
            stderr.WriteLine();
            stderr.WriteLine(rm.GetString("URL") + uri);                
            stderr.WriteLine(rm.GetString("Line") + xe.LineNumber);
            stderr.WriteLine(rm.GetString("Column") + xe.LinePosition);
            stderr.WriteLine(xe);
        }                            
        
        internal static Type FindType(NXsltOptions options, ResourceManager rm, string typeName) {
            Type type;
            Assembly resolverAssembly;
            TextWriter stderr = Console.Error;
            if (options.AssemblyFileName != null) {
                //
                //Load assembly from file
                //                
                try {
                    resolverAssembly = Assembly.LoadFrom(options.AssemblyFileName);
                } catch {
                    //Assembly cannot be loaded
                    stderr.WriteLine();
                    stderr.WriteLine(rm.GetString("Error.Load.Assembly"), options.AssemblyFileName);
                    stderr.WriteLine();
                    return null;
                }
                try {                                
                    type = resolverAssembly.GetType(typeName, true);                            
                } catch {
                    //Type not found in the specified assembly
                    stderr.WriteLine();
                    stderr.WriteLine(rm.GetString("Error.Get.Type.From.Assembly"), 
                        typeName, options.AssemblyFileName);
                    stderr.WriteLine();
                    return null; 
                }                                                                                
            } else if (options.AssemblyName != null) {
                //
                //Load assembly by name
                //                
                try {
                    resolverAssembly = Assembly.LoadWithPartialName(options.AssemblyName);
                } catch {
                    //Assembly cannot be loaded
                    stderr.WriteLine();
                    stderr.WriteLine(rm.GetString("Error.Load.Assembly"), options.AssemblyFileName);
                    stderr.WriteLine();
                    return null;
                }
                try {                                                            
                    type = resolverAssembly.GetType(typeName, true);                            
                } catch {
                    //Type not found
                    stderr.WriteLine();
                    stderr.WriteLine(rm.GetString("Error.Get.Type"), typeName);
                    stderr.WriteLine();
                    return null; 
                }                                                    
            } else {
                //
                //Assembly not specified                        
                //
                try {                                                            
                    type = Type.GetType(typeName, true);                              
                } catch {
                    //Type not found
                    stderr.WriteLine();
                    stderr.WriteLine(rm.GetString("Error.Get.Type"), typeName);
                    stderr.WriteLine();
                    return null; 
                }                                                    
            }
            return type;
        } // FindType()        
    } // NXslt class
} // NXslt namespace
