using System;
using System.Resources;
using System.Xml;
using System.Collections;
using System.Xml.Xsl;

namespace NXslt {
	///	<summary>
	///	Command	line arguments parser.
	/// <p>Author: Oleg Tkachenko, <address>olegt@multiconn.com</address></p>
	///	</summary>
	public class NXsltArgumentsParser {
				
		///	<summary>
		///	Parses command line	arguments into NXsltOptions collection.
		///	</summary>
		///	<param name="args">Command line	arguments.</param>
		/// <param name="rm">Resource manager for L10n.</param>
		///	<returns>Parsed	collection of options or null if the parsing fails.</returns>
		public static NXsltOptions ParseArguments(string[] args, ResourceManager rm) {
			NXsltOptions options = new NXsltOptions();
			XmlNamespaceManager	nm = null;			  
			NameTable nt = null;
			Queue paramQueue = null;			
			for	(int i=0; i<args.Length; i++) {
				string arg = args[i];
				switch (arg) {
				    //Show help
					case "-?":
						options.ShowHelp = true;
						break;
					//Strip whitespace
					case "-xw":
						options.StripWhiteSpace	= true;
						break;
					//Output filename
					case "-o":						  
						if (i == args.Length-1)	{
						    //Absent file name
							NXslt.PrintCommandLineParsingError(rm, "Error.Missing.OutFileName");
							return null;
						} else {
						    //Next argument must be filename
							options.OutFile	= args[++i];
							break;
						}
					case "-exslt":						  
						if (i == args.Length-1)	{
							//Absent exslt lib path
							NXslt.PrintCommandLineParsingError(rm, "Error.Missing.ExsltLibPath");
							return null;
						} else {
							//Next argument must be lib path
							options.ExsltLibPath = args[++i];
							break;
						}
					//Show timings
					case "-t":
					    options.ShowTiming = true;
					    break;
					//Get stylesheet URI from xml-stylesheet PI
					case "-pi":
					    if (options.Stylesheet != null || options.LoadStylesheetFromStdin) {
					        //Both -pi and stylesheet cannot be specified
					        NXslt.PrintCommandLineParsingError(rm, "Error.Both.Stylesheet.And.PI");
					        return null;
					    } else
					        options.GetStylesheetFromPI = true;
					    break;
					//Use Use named URI resolver type
					case "-r":
                        if (i == args.Length-1)	{
                            //Absent resolver type name
                            NXslt.PrintCommandLineParsingError(rm, "Error.Missing.ResolverTypeName");
                            return null;
                        } else {
                            //Next argument must be resolver type
                            options.ResolverTypeName = args[++i];
                            break;
                        }
                    //Assembly file name
                    case "-af":
                        if (options.AssemblyName != null) {
                            NXslt.PrintCommandLineParsingError(rm, "Error.Both.AssemblyFileName.And.Name");
                            return null;
                        }
                        if (i == args.Length-1)	{
                            //Absent assembly file name
                            NXslt.PrintCommandLineParsingError(rm, "Error.Missing.AssemblyFileName");
                            return null;
                        } else {
                            //Next argument must be assembly file name
                            options.AssemblyFileName = args[++i];
                            break;
                        }
                     //Assembly name
                    case "-an":
                        if (options.AssemblyFileName != null) {
                            NXslt.PrintCommandLineParsingError(rm, "Error.Both.AssemblyFileName.And.Name");
                            return null;
                        }
                        if (i == args.Length-1)	{
                            //Absent assembly name
                            NXslt.PrintCommandLineParsingError(rm, "Error.Missing.AssemblyName");
                            return null;
                        } else {
                            //Next argument must be assembly name
                            options.AssemblyName = args[++i];
                            break;
                        }                            
                    //Allow multiple output documents
                    case "-mo":
                        options.MultiOutput = true;
                        break;            					    
					//Validate documents
					case "-v":
					    options.ValidateDocs = true;
					    break;
					//Do not resolve externals
					case "-xe":
					    options.ResolveExternals = false;
					    break;
					//Do not process XInclude
					case "-xi":
						options.ProcessXInclude = false;
						break;
					//Extension class names
					case "-ext":					    
                        if (i == args.Length-1)	{
                            //Absent class names
                            NXslt.PrintCommandLineParsingError(rm, "Error.Missing.ExtClassNames");
                            return null;
                        } else {
                            //Next argument must be ext class names list
                            options.ExtClasses = args[++i].Split(',');
                            break;
                        }                        
					//Source or Stylesheet to be read from stdin
					case "-":
					    if (options.Source == null && !options.LoadSourceFromStdin)
					        options.LoadSourceFromStdin = true;
					    else if (options.Stylesheet == null && !options.LoadStylesheetFromStdin) {
					        //Check out that both source and stylesheet are not "-"
					        if (options.LoadSourceFromStdin) {
					            //Both source and stylesheet cannot be read from stdin
                                NXslt.PrintCommandLineParsingError(rm, "Error.Both.Stdin");
                                return null;
					        } else
					            options.LoadStylesheetFromStdin = true;					    					        
					    } else {
					        //Unrecognized "-"
					        NXslt.PrintCommandLineParsingError(rm);
                            Console.Error.WriteLine(rm.GetString("Error.Unrecognized.Option"), arg);
                            return null;
					    }    
					    break;
					//Other argment - source URI, stylesheet URI, parameter or namespace declaration
					default:
					    //Source URI?
						if (options.Source == null && !options.LoadSourceFromStdin)
							options.Source = arg;
						//Stylesheet URI?
						else if	(options.Stylesheet	== null && !options.LoadStylesheetFromStdin)
							options.Stylesheet = arg;
						//namespace declaration?
						else if	(arg.StartsWith("xmlns:") || arg.StartsWith("xmlns="))	{
						    if (arg.EndsWith("=")) {
						        NXslt.PrintCommandLineParsingError(rm);
						        Console.Error.WriteLine(rm.GetString("Error.Missing.URI"), arg.Remove(arg.Length-1, 1));
						        return null;
						    }
							int	eqIndex = arg.IndexOf('=');
							//Lazy XmlNamespaceManager
							if (nm == null)	{
								nt = new NameTable();
								nm = new XmlNamespaceManager(nt);
							}
							//qname is xmlns:prefix or xmlns
							string qname = arg.Substring(0, eqIndex);
							int	colonIndex = qname.IndexOf(':');
							if (colonIndex != -1)
							    //xmlns:prefix="value" case
								nm.AddNamespace(nt.Add(qname.Substring(colonIndex+1)),
									arg.Substring(eqIndex+1));
							else
							    //xmlns="value" case - default namespace
								nm.AddNamespace(String.Empty,
									arg.Substring(eqIndex+1));
						} 
						//Parameter?
						else if (arg.StartsWith("=")) {
						    NXslt.PrintCommandLineParsingError(rm, "Error.Missing.Name");
						    return null;
						} else if (arg.EndsWith("=")) {
						    NXslt.PrintCommandLineParsingError(rm);
						    Console.Error.WriteLine(rm.GetString("Error.Missing.Value"), arg.Remove(arg.Length-1, 1));
						    return null;
						} else if (arg.IndexOf('=') != -1) {
						    //Parameter - put to the queue till all namespaces are not processed
						    //Lazy param queue
							if (paramQueue == null)
								//Creates queue	with max capacity =	max	possible number	of params
								paramQueue = new Queue(args.Length-2);
							paramQueue.Enqueue(arg);
						} 
						//Unrecognized argument
						else {						    
							NXslt.PrintCommandLineParsingError(rm);
							Console.Error.WriteLine(rm.GetString("Error.Wrong.Param"), arg);
							return null;
						}						 
						break;
				}				 
			}  
			//Resolve parameters
			if (paramQueue != null)	{
				options.XslArgList = new XsltArgumentList();
				foreach	(string	param in paramQueue) {
					int	eqIndex = param.IndexOf('=');
					//qname is prefix:localname or localname
					string qname = param.Substring(0, eqIndex);
					int	colonIndex = qname.IndexOf(':');
					if (colonIndex != -1) {												   
					    //prefix:localname="value" case
						string prefix =	qname.Substring(0, colonIndex);
						string uri;						   
						//Resolve prefix
						if (nm == null || 
							(uri = nm.LookupNamespace(nt.Get(prefix))) == null)	{							 
							NXslt.PrintCommandLineParsingError(rm);
							Console.Error.WriteLine(rm.GetString("Error.Unbounded.Prefix"), prefix);
							return null;
						} 
						else
							options.XslArgList.AddParam(param.Substring(colonIndex+1,	
							    eqIndex-colonIndex-1), uri, param.Substring(eqIndex+1));
					} 
					else 
					    //localname="value"	case - possible default namespace
						options.XslArgList.AddParam(qname, nm==null? String.Empty : nm.DefaultNamespace, 
						    param.Substring(eqIndex+1));
				}  							   
			}
            //Instantiate specified extension objects
            if (options.ExtClasses != null) {
                if (options.XslArgList == null)
                    options.XslArgList = new XsltArgumentList();
                foreach (string typeDecl in options.ExtClasses) {                        
                    string[] parts = typeDecl.Split(new char[]{':'},2);
                    if (parts.Length == 1) {                        
                        Type extObjType = NXslt.FindType(options, rm, parts[0]);
                        if (extObjType == null)
                            return null;
                        try {
                            object o = extObjType.GetConstructor(Type.EmptyTypes).Invoke(null);                            
                            if (nm == null || nm.DefaultNamespace == String.Empty) {
                                //Extension object not bound to a namespace
                                Console.Error.WriteLine();
                                Console.Error.WriteLine(rm.GetString("Error.Ext.No.Namespace"), parts[0]);
                                Console.Error.WriteLine();
                                return null;
                            }
                            string ns = nm.DefaultNamespace;
                            if (options.XslArgList.GetExtensionObject(ns) != null) {
                                //More than one extension object in the same namespace URI
                                Console.Error.WriteLine();
                                Console.Error.WriteLine(rm.GetString("Error.Ext.Namespace.Clash"), ns);
                                Console.Error.WriteLine();
                                return null;
                            }
                            options.XslArgList.AddExtensionObject(ns, o);
                        } catch (Exception e) {
                            //Type cannot be instantiated
                            Console.Error.WriteLine();
                            Console.Error.WriteLine(rm.GetString("Error.Create.Resolver"), parts[0], e);
                            Console.Error.WriteLine();
                            return null;                                                    
                        }
                    } else {
                        string prefix = parts[0];
                        string uri;
                        //Resolve prefix
                        if (nm == null || 
                            (uri = nm.LookupNamespace(nt.Get(prefix))) == null)	{							 
                            NXslt.PrintCommandLineParsingError(rm);
                            Console.Error.WriteLine(rm.GetString("Error.Unbounded.Prefix"), prefix);
                            return null;
                        }
                        
                        Type extObjType = NXslt.FindType(options, rm, parts[1]);
                        if (extObjType == null)
                            return null;
                        try {
                            object o = extObjType.GetConstructor(Type.EmptyTypes).Invoke(null);                            
                            if (options.XslArgList.GetExtensionObject(uri) != null) {
                                //More than one extension object in the same namespace URI
                                Console.Error.WriteLine();
                                Console.Error.WriteLine(rm.GetString("Error.Ext.Namespace.Clash"), uri);
                                Console.Error.WriteLine();
                                return null;
                            }
                            options.XslArgList.AddExtensionObject(uri, o);
                        } catch(Exception e) {
                            //Type cannot be instantiated
                            Console.Error.WriteLine();
                            Console.Error.WriteLine(rm.GetString("Error.Create.Resolver"), parts[1], e);
                            Console.Error.WriteLine();
                            return null;                                                    
                        }    
                    }                        
                }
            }			
			return options;	
		} // ParseArguments
    } // NXsltArgumentsParser 
} // NXslt
